@@ -10,10 +10,13 @@ import contextlib
 import functools
 import importlib
 import inspect
+import itertools
 import re
 import pydoc
 import sys
+import types
 import typing
+import warnings
 from argparse import (
     REMAINDER, SUPPRESS,
     Action, ArgumentParser, RawTextHelpFormatter,
@@ -21,7 +24,20 @@ from argparse import (
 from collections import defaultdict, namedtuple, Counter
 from enum import Enum
 from pathlib import PurePath
-from typing import Any, Callable, Dict, List, Optional, Union
+from typing import Any, Callable, Dict, List, Optional, Tuple, Union
+
+try:
+    import importlib.metadata as _im
+except ImportError:
+    import importlib_metadata as _im
+try:
+    from typing import Annotated
+except ImportError:
+    from typing_extensions import Annotated
+try:
+    from typing import Literal
+except ImportError:
+    from typing_extensions import Literal
 
 import docutils.core
 from docutils.nodes import NodeVisitor, SkipNode, TextElement
@@ -34,43 +50,35 @@ finally:
     if sys.version_info >= (3, 7):
         del collections.Callable
 
-try:
-    from typing import Literal
-except ImportError:
-    from typing_extensions import Literal
-
 try:
     # colorama is a dependency on Windows to support ANSI escapes (from rst
-    # markup).  It is optional on Unices, but can still useful be there as it
+    # markup).  It is optional on Unices, but can still be useful there as it
     # strips out ANSI escapes when the output is piped.
     from colorama import colorama_text as _colorama_text
 except ImportError:
-    @contextlib.contextmanager
-    def _colorama_text(*args):
-        yield
+    _colorama_text = getattr(contextlib, 'nullcontext', contextlib.ExitStack)
 
 try:
-    from _defopt_version import version as __version__
+    __version__ = _im.version('defopt')
 except ImportError:
-    pass
+    __version__ = '0+unknown'
 
-__all__ = ['run', 'signature']
+__all__ = ['run', 'signature', 'bind']
 
 _PARAM_TYPES = ['param', 'parameter', 'arg', 'argument', 'key', 'keyword']
 _TYPE_NAMES = ['type', 'kwtype']
 
 _Doc = namedtuple('_Doc', ('first_line', 'text', 'params', 'raises'))
 _Param = namedtuple('_Param', ('text', 'type'))
+class _Raises(tuple): pass
 
 
 if hasattr(typing, 'get_args'):
     _ti_get_args = typing.get_args
 else:
-    def _ti_get_args(tp):
-        import typing_inspect as ti
-        if type(tp) is type(Literal):  # Py<3.7.
-            return tp.__values__
-        return ti.get_args(tp, evaluate=True)  # =True is default on Py>=3.7.
+    import typing_inspect as _ti
+    # evaluate=True is default on Py>=3.7.
+    _ti_get_args = functools.partial(_ti.get_args, evaluate=True)
 
 
 if hasattr(typing, 'get_origin'):
@@ -78,12 +86,14 @@ if hasattr(typing, 'get_origin'):
 else:
     def _ti_get_origin(tp):
         import typing_inspect as ti
-        if type(tp) is type(Literal):  # Py<3.7.
+        if ti.is_literal_type(tp):  # ti.get_origin returns None for Literals.
             return Literal
         origin = ti.get_origin(tp)
         return {  # Py<3.7.
             typing.List: list,
             typing.Iterable: collections.abc.Iterable,
+            getattr(typing, 'Collection', object()):
+                getattr(collections.abc, 'Collection', object()),
             typing.Sequence: collections.abc.Sequence,
             typing.Tuple: tuple,
         }.get(origin, origin)
@@ -129,10 +139,63 @@ class _DefaultList(list):
     """
 
 
+_unset = 'UNSET'
+
+
+def bind(funcs: Union[Callable, List[Callable], Dict[str, Callable]], *,
+         parsers: Dict[type, Callable[[str], Any]] = {},
+         short: Optional[Dict[str, str]] = None,
+         cli_options: Literal['kwonly', 'all', 'has_default'] = _unset,
+         strict_kwonly=_unset,
+         show_defaults: bool = True,
+         show_types: bool = False,
+         no_negated_flags: bool = False,
+         version: Union[str, None, bool] = None,
+         argparse_kwargs: dict = {},
+         argv: Optional[List[str]] = None):
+    """
+    Process command-line arguments and bind arguments.
+
+    This function takes the same parameters as `defopt.run`, but returns a
+    pair, which consists of a `~typing.Callable` *func* and a
+    `~inspect.BoundArguments` *ba*, such that `defopt.run` would call
+    ``func(*ba.args, **ba.kwargs)`` (modulo exception handling).
+    """
+    if strict_kwonly == _unset:
+        if cli_options == _unset:
+            cli_options = 'kwonly'
+    else:
+        if cli_options != _unset:
+            raise ValueError(
+                "Cannot pass both 'cli_options' and 'strict_kwonly'")
+        warnings.warn(
+            'strict_kwonly is deprecated and will be removed in an upcoming '
+            'release', DeprecationWarning)
+        cli_options = 'kwonly' if strict_kwonly else 'has_default'
+    parser = _create_parser(
+        funcs, parsers=parsers, short=short, cli_options=cli_options,
+        show_defaults=show_defaults, show_types=show_types,
+        no_negated_flags=no_negated_flags, version=version,
+        argparse_kwargs=argparse_kwargs)
+    with _colorama_text():
+        parsed_argv = vars(parser.parse_args(argv))
+    try:
+        func = parsed_argv.pop('_func')
+    except KeyError:
+        # Workaround for http://bugs.python.org/issue9253#msg186387 (and
+        # https://bugs.python.org/issue29298 which blocks using required=True).
+        parser.error('too few arguments')
+    sig = signature(func)
+    ba = sig.bind_partial()
+    ba.arguments.update(parsed_argv)
+    return func, ba
+
+
 def run(funcs: Union[Callable, List[Callable], Dict[str, Callable]], *,
         parsers: Dict[type, Callable[[str], Any]] = {},
         short: Optional[Dict[str, str]] = None,
-        strict_kwonly: bool = True,
+        cli_options: Literal['kwonly', 'all', 'has_default'] = _unset,
+        strict_kwonly=_unset,
         show_defaults: bool = True,
         show_types: bool = False,
         no_negated_flags: bool = False,
@@ -140,7 +203,7 @@ def run(funcs: Union[Callable, List[Callable], Dict[str, Callable]], *,
         argparse_kwargs: dict = {},
         argv: Optional[List[str]] = None):
     """
-    Process command line arguments and run the given functions.
+    Process command-line arguments and run the given functions.
 
     ``funcs`` can be a single callable, which is parsed and run; or it can
     be a list of callables or mappable of strs to callables, in which case
@@ -156,16 +219,22 @@ def run(funcs: Union[Callable, List[Callable], Dict[str, Callable]], *,
     :param short:
         Dictionary mapping parameter names (after conversion of underscores to
         dashes) to letters, to use as alternative short flags.  Defaults to
-        ``None``, which means to generate short flags for any non-ambiguous
+        `None`, which means to generate short flags for any non-ambiguous
         option.  Set to ``{}`` to completely disable short flags.
+    :param cli_options:
+        The default behavior ('kwonly') is to convert keyword-only parameters
+        to command line flags, and non-keyword-only parameters with a default
+        to optional positional command line parameters. 'all' turns all
+        parameters into command-line flags. 'has_default' turns a parameter
+        into a command-line flag if and only if it has a default value.
     :param strict_kwonly:
-        If `False`, all parameters with a default are converted into
-        command-line flags.  The default behavior (`True`) is to convert
+        Deprecated.  If `False`, all parameters with a default are converted
+        into command-line flags. The default behavior (`True`) is to convert
         keyword-only parameters to command line flags, and non-keyword-only
         parameters with a default to optional positional command line
         parameters.
     :param show_defaults:
-        Whether parameter defaults are appneded to parameter descriptions.
+        Whether parameter defaults are appended to parameter descriptions.
     :param show_types:
         Whether parameter types are appended to parameter descriptions.
     :param no_negated_flags:
@@ -175,43 +244,74 @@ def run(funcs: Union[Callable, List[Callable], Dict[str, Callable]], *,
     :param version:
         If a string, add a ``--version`` flag which prints the given version
         string and exits.
-        If ``True``, the version string is auto-detected by searching for a
+        If `True`, the version string is auto-detected by searching for a
         ``__version__`` attribute on the module where the function is defined,
         and its parent packages, if any.  Error out if such a version cannot be
         found, or if multiple callables with different version strings are
         passed.
-        If ``None`` (the default), behave as for ``True``, but don't add a
+        If `None` (the default), behave as for `True`, but don't add a
         ``--version`` flag if no version string can be autodetected.
-        If ``False``, do not add a ``--version`` flag.
+        If `False`, do not add a ``--version`` flag.
     :param argparse_kwargs:
         A mapping of keyword arguments that will be passed to the
-        ArgumentParser constructor.
+        `~argparse.ArgumentParser` constructor.
     :param argv:
         Command line arguments to parse (default: ``sys.argv[1:]``).
     :return:
         The value returned by the function that was run.
     """
-    parser = _create_parser(
-        funcs, parsers=parsers, short=short, strict_kwonly=strict_kwonly,
-        show_defaults=show_defaults, show_types=show_types,
-        no_negated_flags=no_negated_flags,
-        version=version, argparse_kwargs=argparse_kwargs)
-    with _colorama_text():
-        args = parser.parse_args(argv)
-    # Workaround for http://bugs.python.org/issue9253#msg186387
-    if not hasattr(args, '_func'):
-        parser.error('too few arguments')
+    func, ba = bind(
+        funcs, parsers=parsers, short=short, cli_options=cli_options,
+        strict_kwonly=strict_kwonly, show_defaults=show_defaults,
+        show_types=show_types, no_negated_flags=no_negated_flags,
+        version=version, argparse_kwargs=argparse_kwargs, argv=argv)
+    sig = signature(func)
+    raises, = [
+        # typing_inspect does not allow fetching metadata; see e.g. ti#82.
+        arg for arg in getattr(sig.return_annotation, '__metadata__', [])
+        if isinstance(arg, _Raises)]
+    # The function call should occur here to minimize effects on the traceback.
     try:
-        return _call_function(parser, args._func, args)
-    except args._exc_types as e:
+        return func(*ba.args, **ba.kwargs)
+    except raises as e:
         sys.exit(e)
 
 
+def _recurse_functions(funcs, subparsers):
+    if not isinstance(funcs, collections.abc.Mapping):
+        # If this iterable is not a maping, then convert it to one using the
+        # function name itself as the key, but replacing _ with -.
+        try:
+            funcs = {func.__name__.replace('_', '-'): func for func in funcs}
+        except AttributeError as exc:
+            # Do not allow a mapping inside of a list
+            raise ValueError(
+                'Use dictionaries (mappings) for nesting; other iterables may '
+                'only contain functions (callables).'
+            ) from exc
+
+    for name, func in funcs.items():
+        if callable(func):
+            # If this item is callable, then add it to the current
+            # subparser using this name.
+            subparser = subparsers.add_parser(
+                name,
+                formatter_class=RawTextHelpFormatter,
+                help=_parse_docstring(inspect.getdoc(func)).first_line)
+            yield func, subparser
+        else:
+            # If this item is not callable, then add this name as a new
+            # subparser and recurse the the items.
+            nestedsubparser = subparsers.add_parser(name)
+            nestedsubparsers = nestedsubparser.add_subparsers()
+            yield from _recurse_functions(func, nestedsubparsers)
+
+
 def _create_parser(
         funcs, *,
         parsers={},
         short=None,
-        strict_kwonly=True,
+        cli_options='kwonly',
         show_defaults=True,
         show_types=False,
         no_negated_flags=False,
@@ -221,21 +321,13 @@ def _create_parser(
         **{**{'formatter_class': RawTextHelpFormatter}, **argparse_kwargs})
     version_sources = []
     if callable(funcs):
-        _populate_parser(funcs, parser, parsers, short, strict_kwonly,
+        _populate_parser(funcs, parser, parsers, short, cli_options,
                          show_defaults, show_types, no_negated_flags)
         version_sources.append(funcs)
     else:
         subparsers = parser.add_subparsers()
-        for func in funcs:
-            if isinstance(funcs, collections.abc.MutableMapping):
-                name, func = func, funcs[func]
-            else:
-                name = func.__name__.replace('_', '-')
-            subparser = subparsers.add_parser(
-                name,
-                formatter_class=RawTextHelpFormatter,
-                help=_parse_docstring(inspect.getdoc(func)).first_line)
-            _populate_parser(func, subparser, parsers, short, strict_kwonly,
+        for func, subparser in _recurse_functions(funcs, subparsers):
+            _populate_parser(func, subparser, parsers, short, cli_options,
                              show_defaults, show_types, no_negated_flags)
             version_sources.append(func)
     if isinstance(version, str):
@@ -259,9 +351,8 @@ def _get_version(funcs):
 
 
 def _get_version1(func):
-    try:
-        module_name = func.__module__
-    except AttributeError:
+    module_name = getattr(func, '__module__', None)
+    if not module_name:
         return
     if module_name == '__main__':
         f_globals = getattr(func, '__globals__', {})
@@ -300,42 +391,57 @@ def signature(func: Callable):
     differences:
 
     - Private parameters (starting with an underscore) are not listed.
-    - Parameter annotations are also read from ``func``'s docstring (if a
-      parameter's type is specified both in the signature and the docstring,
-      both types must match).
+    - Parameter types are also read from ``func``'s docstring (if a parameter's
+      type is specified both in the signature and the docstring, both types
+      must match).
     - The docstring for each parameter is available as the
       `~inspect.Parameter`'s ``.doc`` attribute (in fact, a subclass of
       `~inspect.Parameter` is used).
+    - The return type is `~typing.Annotated` with the documented raisable
+      exception types, in wrapped in a private tuple subclass.
     """
-    full_sig = inspect.signature(func)
+    return _signature(func)
+
+
+def _signature(func, *, skip_first_arg=False):
+    # See _is_constructible_from_str for skip_first_arg.  We can't just drop
+    # the first arg later because it may be unannotated.
+    orig_sig = inspect.signature(func)
+    orig_params = orig_sig.parameters.values()
+    if skip_first_arg:
+        _, *orig_params = orig_params
     doc = _parse_docstring(inspect.getdoc(func))
     parameters = []
-    for param in full_sig.parameters.values():
+    for param in orig_params:
         if param.name.startswith('_'):
             if param.default is param.empty:
                 raise ValueError(
                     'Parameter {} of {}{} is private but has no default'
-                    .format(param.name, func.__name__, full_sig))
+                    .format(param.name, func.__name__, orig_sig))
         else:
             parameters.append(Parameter(
                 name=param.name, kind=param.kind, default=param.default,
                 annotation=_get_type(func, param.name),
                 doc=doc.params.get(param.name, _Param(None, None)).text))
-    return full_sig.replace(parameters=parameters)
+    exc_types = _Raises(_get_type_from_doc(name, func.__globals__)
+                        for name in doc.raises)
+    return_annotation = Annotated[orig_sig.return_annotation, exc_types]
+    return orig_sig.replace(
+        parameters=parameters, return_annotation=return_annotation)
 
 
-def _populate_parser(func, parser, parsers, short, strict_kwonly,
+def _populate_parser(func, parser, parsers, short, cli_options,
                      show_defaults, show_types, no_negated_flags):
     sig = signature(func)
     doc = _parse_docstring(inspect.getdoc(func))
     parser.description = doc.text
 
-    exc_types = tuple(_get_type_from_doc(name, func.__globals__)
-                      for name in doc.raises)
-    positionals = {name for name, param in sig.parameters.items()
-                   if ((param.default is param.empty or strict_kwonly)
-                       and not _is_list_like(param.annotation)
-                       and param.kind != param.KEYWORD_ONLY)}
+    positionals = {
+        name for name, param in sig.parameters.items()
+        if ((cli_options == 'kwonly' or
+             (param.default is param.empty and cli_options == 'has_default'))
+            and not _is_list_like(param.annotation)
+            and param.kind != param.KEYWORD_ONLY)}
     if short is None:
         count_initials = Counter(name[0] for name in sig.parameters
                                  if name not in positionals)
@@ -356,23 +462,14 @@ def _populate_parser(func, parser, parsers, short, strict_kwonly,
         default = param.default if hasdefault else SUPPRESS
         required = not hasdefault and param.kind != param.VAR_POSITIONAL
         positional = name in positionals
-        if type_ == bool and not positional:
-            if no_negated_flags and default is False:
-                actions.append(
-                    _add_argument(parser, name, short,
-                                  action='store_true',
-                                  default=default,
-                                  # Add help if available.
-                                  **kwargs))
-            else:
-                # Special case:
-                # just add parameterless --name and --no-name flags.
-                actions.append(_add_argument(parser, name, short,
-                                             action=_BooleanOptionalAction,
-                                             default=default,
-                                             required=required,
-                                             # Add help if available.
-                                             **kwargs))
+        if type_ in [bool, typing.Optional[bool]] and not positional:
+            action = ('store_true' if no_negated_flags and
+                      default in [False, None]
+                      else _BooleanOptionalAction)  # --name/--no-name
+            actions.append(_add_argument(
+                parser, name, short, action=action, default=default,
+                required=required,  # Always False if `default is False`.
+                **kwargs))  # Add help if available.
             continue
         # Always set a default, even for required parameters, so that we can
         # later (ab)use default == SUPPRESS (!= None) to detect required
@@ -396,9 +493,16 @@ def _populate_parser(func, parser, parsers, short, strict_kwonly,
         member_types = None
         if _ti_get_origin(type_) is tuple:
             member_types = _ti_get_args(type_)
-            kwargs['nargs'] = len(member_types)
-            kwargs['action'] = _make_store_tuple_action_class(
-                tuple, member_types, parsers)
+            # Variable-length tuples of homogenous type are specified like
+            # Tuple[int, ...]
+            if len(member_types) == 2 and member_types[1] is Ellipsis:
+                kwargs['nargs'] = "*"
+                kwargs['action'] = _make_store_tuple_action_class(
+                    tuple, member_types, parsers, is_variable_length=True)
+            else:
+                kwargs['nargs'] = len(member_types)
+                kwargs['action'] = _make_store_tuple_action_class(
+                    tuple, member_types, parsers)
         elif (isinstance(type_, type) and issubclass(type_, tuple)
               and hasattr(type_, '_fields')):
             # Before Py3.6, `_field_types` does not preserve order, so retrieve
@@ -407,8 +511,7 @@ def _populate_parser(func, parser, parsers, short, strict_kwonly,
             member_types = tuple(hints[field] for field in type_._fields)
             kwargs['nargs'] = len(member_types)
             kwargs['action'] = _make_store_tuple_action_class(
-                lambda args, type_=type_: type_(*args),
-                member_types, parsers)
+                type_, member_types, parsers)
             if not positional:  # http://bugs.python.org/issue14074
                 kwargs['metavar'] = type_._fields
         else:
@@ -423,7 +526,7 @@ def _populate_parser(func, parser, parsers, short, strict_kwonly,
         _update_help_string(
             action, show_defaults=show_defaults, show_types=show_types)
 
-    parser.set_defaults(_func=func, _exc_types=exc_types)
+    parser.set_defaults(_func=func)
 
 
 def _add_argument(parser, name, short, _positional=False, **kwargs):
@@ -474,9 +577,12 @@ def _update_help_string(action, *, show_defaults, show_types):
 
 
 def _is_list_like(type_):
-    return (_ti_get_origin(type_)
-            in [list, collections.abc.Iterable, collections.abc.Sequence])
-
+    return _ti_get_origin(type_) in [
+        list,
+        collections.abc.Iterable,
+        getattr(collections.abc, 'Collection', object()),
+        collections.abc.Sequence,
+    ]
 
 def _get_type(func, name):
     """
@@ -495,6 +601,13 @@ def _get_type(func, name):
     except KeyError:
         hint_type = None
     else:
+        param = inspect.signature(func).parameters[name]
+        if (param.default is None
+                and param.annotation != hint
+                and Optional[param.annotation] == hint):
+            # `f(x: tuple[int, int] = None)` means we support a tuple, but not
+            # None (to constrain the number of arguments).
+            hint = param.annotation
         hint_type = _get_type_from_hint(hint)
 
     chosen = [x is not None for x in [doc_type, hint_type]]
@@ -514,15 +627,14 @@ def _get_type_from_doc(name, globalns):
             raise ValueError(
                 'unsupported union including container type: {}'.format(name))
         return Union[tuple(subtype for subtype in subtypes)]
-    # Support for legacy list syntax "list[type]".
-    # (This intentionally won't catch `List` or `typing.List`)
-    match = re.match(r'([a-z]\w+)\[([\w\.]+)\]', name)
+    # Support for sphinx-specific "list[type]", "tuple[type]" syntax; only
+    # needed for Py<3.9.
+    # (This intentionally won't catch `List` or `typing.List`.)
+    match = re.match(r'(list|tuple)\[([\w\.]+)\]', name)
     if match:
         container, type_ = match.groups()
-        if container != 'list':
-            raise ValueError(
-                'unsupported container type: {}'.format(container))
-        return List[eval(type_, globalns)]
+        container = {'list': List, 'tuple': Tuple}[container]
+        return container[eval(type_, globalns)]
     return _get_type_from_hint(eval(name, globalns))
 
 
@@ -542,21 +654,6 @@ def _get_type_from_hint(hint):
     return hint
 
 
-def _call_function(parser, func, args):
-    positionals = []
-    keywords = {}
-    sig = signature(func)
-    for name, param in sig.parameters.items():
-        arg = getattr(args, name)
-        if param.kind in [param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD]:
-            positionals.append(arg)
-        elif param.kind == param.VAR_POSITIONAL:
-            positionals.extend(arg)
-        else:
-            keywords[name] = arg
-    return func(*positionals, **keywords)
-
-
 def _passthrough_role(
         name, rawtext, text, lineno, inliner, options={}, content=[]):
     return [TextElement(rawtext, text)], []
@@ -595,7 +692,11 @@ def _parse_docstring(doc):
     doc = str(NumpyDocstring(doc, cfg))
 
     with _sphinx_common_roles():
-        tree = docutils.core.publish_doctree(doc)
+        tree = docutils.core.publish_doctree(
+            # Disable syntax highlighting, as 1) pygments is not a dependency
+            # 2) we don't render with colors and 3) SH breaks the assumption
+            # that literal blocks contain a single text element.
+            doc, settings_overrides={'syntax_highlight': 'none'})
 
     class Visitor(NodeVisitor):
         optional = [
@@ -662,7 +763,7 @@ def _parse_docstring(doc):
         def depart_rubric(self, node):
             # Style consistent with "usage:", "positional arguments:", etc.
             self._current_paragraph[:] = [
-                (t.lower() if t == t.title() else t) + ":"
+                (t.lower() if t == t.title() else t) + ':'
                 for t in self._current_paragraph]
             self.depart_paragraph(node)
 
@@ -759,11 +860,16 @@ def _parse_docstring(doc):
             if paragraph is comment_token:
                 continue
             text.append(paragraph)
-            # We insert a space before each newline to prevent argparse
-            # from stripping consecutive newlines down to just two
-            # (http://bugs.python.org/issue31330).
-            # FIXME: but napoleon inserts far too many newlines :/
-            text.append(' \n' * (next_start - start - paragraph.count('\n')))
+            # Insert two newlines to separate paragraphs by a blank line.
+            # Actually, paragraphs may or may not already have a trailing
+            # newline (e.g. text paragraphs do but literal blocks don't) but
+            # argparse will strip extra newlines anyways.  This means that
+            # extra blank lines in the original docstring will be stripped, but
+            # this is less ugly than having a large number of extra blank lines
+            # arising e.g. from skipped info fields (which are not rendered).
+            # This means that list items are always separated by blank lines,
+            # which is an acceptable tradeoff for now.
+            text.append('\n\n')
         parsed = _Doc(text[0], ''.join(text), tuples, visitor.raises)
     else:
         parsed = _Doc('', '', tuples, visitor.raises)
@@ -771,37 +877,43 @@ def _parse_docstring(doc):
 
 
 def _get_parser(type_, parsers):
-    try:
+    if type_ in parsers:  # Not catching KeyError, to avoid exception chaining.
         parser = functools.partial(parsers[type_])
-    except KeyError:
-        if (type_ in [str, int, float]
-                or isinstance(type_, type) and issubclass(type_, PurePath)):
-            parser = functools.partial(type_)
-        elif type_ == bool:
-            parser = functools.partial(_parse_bool)
-        elif type_ == slice:
-            parser = functools.partial(_parse_slice)
-        elif type_ == type(None):
-            parser = functools.partial(_parse_none)
-        elif type_ == list:
-            raise ValueError('unable to parse list (try list[type])')
-        elif isinstance(type_, type) and issubclass(type_, Enum):
-            parser = _make_enum_parser(type_)
-        elif _is_constructible_from_str(type_):
-            parser = functools.partial(type_)
-        elif _ti_get_origin(type_) is Union:
-            parser = _make_union_parser(
-                type_,
-                [_get_parser(arg, parsers) for arg in _ti_get_args(type_)])
-        elif _ti_get_origin(type_) is Literal:  # Py>=3.7.
-            parser = _make_literal_parser(
-                type_,
-                [_get_parser(type(arg), parsers)
-                 for arg in _ti_get_args(type_)])
-        else:
-            raise Exception('no parser found for type {}'.format(
-                # typing types have no __name__.
-                getattr(type_, '__name__', repr(type_))))
+    elif (type_ in [str, int, float]
+          or isinstance(type_, type) and issubclass(type_, PurePath)):
+        parser = functools.partial(type_)
+    elif type_ == bool:
+        parser = functools.partial(_parse_bool)
+    elif type_ == slice:
+        parser = functools.partial(_parse_slice)
+    elif type_ == type(None):
+        parser = functools.partial(_parse_none)
+    elif type_ == list:
+        raise ValueError('unable to parse list (try list[type])')
+    elif isinstance(type_, type) and issubclass(type_, Enum):
+        parser = _make_enum_parser(type_)
+    elif _is_constructible_from_str(type_):
+        parser = functools.partial(type_)
+    elif _ti_get_origin(type_) in [Union, getattr(types, "UnionType", "")]:
+        args = _ti_get_args(type_)
+        if type(None) in args:
+            # If None is in the Union, parse it first.  This only matters if
+            # there's a custom parser for None, in which case the user should
+            # normally have picked values that they *want* to be parsed as
+            # None as opposed to anything else, e.g. strs, even if that was
+            # possible.
+            args = (type(None),
+                    *[arg for arg in args if arg is not type(None)])
+        parser = _make_union_parser(
+            type_, [_get_parser(arg, parsers) for arg in args])
+    elif _ti_get_origin(type_) is Literal:
+        args = _ti_get_args(type_)
+        parser = _make_literal_parser(
+            type_, [_get_parser(type(arg), parsers) for arg in args])
+    else:
+        raise Exception('no parser found for type {}'.format(
+            # typing types have no __name__.
+            getattr(type_, '__name__', repr(type_))))
     # Set the name that the user expects to see in error messages (we always
     # return a temporary partial object so it's safe to set its __name__).
     # Unions and Literals don't have a __name__, but their str is fine.
@@ -851,9 +963,9 @@ def _make_enum_parser(enum, value=None):
                 value, ', '.join(map(repr, enum.__members__))))
 
 
-def _is_constructible_from_str(type_):
+def _is_constructible_from_str(type_, *, skip_first_arg=False):
     try:
-        sig = signature(type_)
+        sig = _signature(type_, skip_first_arg=skip_first_arg)
         (argname, _), = sig.bind(object()).arguments.items()
     except TypeError:  # Can be raised by signature() or Signature.bind().
         return False
@@ -866,8 +978,11 @@ def _is_constructible_from_str(type_):
             return True
     if isinstance(type_, type):
         # signature() first checks __new__, if it is present.
+        # skip_first_arg behaves as if the first parameter was bound, i.e.
+        # __init__.__get__(object(), type_) but the latter can fail for types
+        # implemented in C (which don't support binding arbitrary objects).
         return _is_constructible_from_str(
-            type_.__init__.__get__(object(), type_))
+            type_.__init__, skip_first_arg=True)
     return False
 
 
@@ -877,7 +992,7 @@ def _make_union_parser(union, parsers, value=None):
     for p in parsers:
         try:
             return p(value)
-        except (ValueError, ArgumentTypeError) as e:
+        except (ValueError, ArgumentTypeError):
             pass
     raise ValueError(
         '{} could not be parsed as any of {}'.format(value, union))
@@ -898,15 +1013,23 @@ def _make_literal_parser(literal, parsers, value=None):
             value, ', '.join(map(repr, map(str, _ti_get_args(literal))))))
 
 
-def _make_store_tuple_action_class(make_tuple, member_types, parsers):
+def _make_store_tuple_action_class(
+    tuple_type, member_types, parsers, *, is_variable_length=False):
+    if is_variable_length:
+        parsers = itertools.repeat(_get_parser(member_types[0], parsers))
+    else:
+        parsers = [_get_parser(arg, parsers) for arg in member_types]
+
     class _StoreTupleAction(Action):
         def __call__(self, parser, namespace, values, option_string=None):
             try:
-                value = make_tuple(_get_parser(arg, parsers)(value)
-                                   for arg, value in zip(member_types, values))
+                value = tuple(p(value) for p, value in zip(parsers, values))
             except ArgumentTypeError as exc:
                 raise ArgumentError(self, str(exc))
+            if tuple_type is not tuple:
+                value = tuple_type(*value)
             setattr(namespace, self.dest, value)
+
     return _StoreTupleAction
 
 
