@@ -4,6 +4,8 @@ import inspect
 import re
 import subprocess
 import sys
+import tokenize
+import types
 import typing
 import unittest
 from contextlib import ExitStack
@@ -42,6 +44,10 @@ class NotConstructibleFromStr:
         pass
 
 
+class EmptyType:
+    pass
+
+
 class TestDefopt(unittest.TestCase):
     def test_main(self):
         def main(foo):
@@ -68,6 +74,133 @@ class TestDefopt(unittest.TestCase):
             defopt.run({"sub1": sub, "sub_2": sub_with_dash},
                        argv=['sub_2', '--baz', '1']), 1)
 
+    def test_nested_lists_invalid(self):
+        def sub1(*bar):
+            """:type bar: float"""
+            return bar
+        def subsub1(*, baz=None):
+            """:type baz: int"""
+            return baz
+        def subsub2(*foo):
+            """:type foo: float"""
+            return foo
+        with self.assertRaises(ValueError):
+            defopt.run([sub1, [subsub1, subsub2]], argv=['sub1', '1.2'])
+
+    def test_nested_subcommands1(self):
+        def sub1(*bar):
+            """:type bar: float"""
+            return bar
+        def subsub1(*, baz=None):
+            """:type baz: int"""
+            return baz
+        def subsub2(*foo):
+            """:type foo: float"""
+            return foo
+        self.assertEqual(
+            defopt.run({"sub-1": [sub1], "sub-2": [subsub1, subsub2]},
+                       argv=['sub-1', 'sub1', '1.2']), (1.2,))
+        self.assertEqual(
+            defopt.run({"sub-1": [sub1], "sub-2": [subsub1, subsub2]},
+                       argv=['sub-2', 'subsub1', '--baz', '1']), 1)
+        self.assertEqual(
+            defopt.run({"sub-1": [sub1], "sub-2": [subsub1, subsub2]},
+                       argv=['sub-2', 'subsub2', '1.5']), (1.5,))
+
+    def test_nested_subcommands2(self):
+        def sub1(*bar):
+            """:type bar: float"""
+            return bar
+        def subsub1(*, baz=None):
+            """:type baz: int"""
+            return baz
+        def subsub2(*foo):
+            """:type foo: float"""
+            return foo
+        self.assertEqual(
+            defopt.run({"sub-1": sub1, "sub-2": [subsub1, subsub2]},
+                       argv=['sub-1', '1.2']), (1.2,))
+        self.assertEqual(
+            defopt.run({"sub1": sub1, "sub-2": [subsub1, subsub2]},
+                       argv=['sub-2', 'subsub1', '--baz', '1']), 1)
+        self.assertEqual(
+            defopt.run({"sub1": sub1, "sub-2": [subsub1, subsub2]},
+                       argv=['sub-2', 'subsub2', '1.5']), (1.5,))
+
+    def test_nested_subcommands3(self):
+        def sub1(*bar):
+            """:type bar: float"""
+            return bar
+        def subsub1(*, baz=None):
+            """:type baz: int"""
+            return baz
+        def subsub2(*foo):
+            """:type foo: float"""
+            return foo
+        self.assertEqual(
+            defopt.run({"sub-1": sub1,
+                        "sub-2": {'subsub1': subsub1, 'subsub2': subsub2}},
+                       argv=['sub-1', '1.2']), (1.2,))
+        self.assertEqual(
+            defopt.run({"sub-1": sub1,
+                        "sub-2": {'subsub1': subsub1, 'subsub2': subsub2}},
+                       argv=['sub-2', 'subsub1', '--baz', '1']), 1)
+        self.assertEqual(
+            defopt.run({"sub-1": sub1,
+                        "sub-2": {'subsub1': subsub1, 'subsub2': subsub2}},
+                       argv=['sub-2', 'subsub2', '1.5']), (1.5,))
+
+    def test_nested_subcommands_deep(self):
+        def sub(*bar):
+            """:type bar: float"""
+            return bar
+        self.assertEqual(
+            defopt.run({'a': {'b': {'c': {'d': {'e': sub}}}}},
+                       argv=['a', 'b', 'c', 'd', 'e', '1.2']), (1.2,))
+        self.assertEqual(
+            defopt.run({'a': {'b': {'c': {'d': {'e': [sub]}}}}},
+                       argv=['a', 'b', 'c', 'd', 'e', 'sub', '1.2']), (1.2,))
+
+    def test_nested_subcommands_mixed_invalid1(self):
+        def sub1(*bar):
+            """:type bar: float"""
+            return bar
+        def subsub1(*, baz=None):
+            """:type baz: int"""
+            return baz
+        def subsub2(*foo):
+            """:type foo: float"""
+            return foo
+        with self.assertRaises(ValueError):
+            defopt.run([sub1, {'sub2': [subsub1, subsub2]}],
+                        argv=['sub1', '1.2'])
+        with self.assertRaises(ValueError):
+                defopt.run([sub1, {'sub2': [subsub1, subsub2]}],
+                        argv=['sub2', 'subsub1', '--baz', '1'])
+        with self.assertRaises(ValueError):
+                defopt.run([sub1, {'sub2': [subsub1, subsub2]}],
+                        argv=['sub2', 'subsub2', '1.1'])
+
+    def test_nested_subcommands_mixed_invalid2(self):
+        def sub(*bar):
+            """:type bar: float"""
+            return bar
+        def subsub_with_dash(*, baz=None):
+            """:type baz: int"""
+            return baz
+        def subsub(*foo):
+            """:type foo: float"""
+            return foo
+        with self.assertRaises(ValueError):
+            defopt.run([sub, {'subsub1': subsub_with_dash, 'subsub2': subsub}],
+                       argv=['sub', '1.2'])
+        with self.assertRaises(ValueError):
+            defopt.run([sub, {'subsub1': subsub_with_dash, 'subsub2': subsub}],
+                       argv=['subsub1', '--baz', '1'])
+        with self.assertRaises(ValueError):
+            defopt.run([sub, {'subsub1': subsub_with_dash, 'subsub2': subsub}],
+                       argv=['subsub2', '1.5'])
+
     def test_var_positional(self):
         for doc in [
                 ":type foo: int", r":type \*foo: int", ":param int foo: doc"]:
@@ -178,7 +311,6 @@ class TestParsers(unittest.TestCase):
     def test_overridden_parser(self):
         def parser(string):
             return int(string) * 2
-
         def main(value):
             """:type value: int"""
             return value
@@ -187,17 +319,20 @@ class TestParsers(unittest.TestCase):
 
     def test_overridden_none_parser(self):
         def parser(string):
-            if string == ':none:':
+            if string == 'nil':
                 return None
             else:
-                raise ValueError("Not :none:")
-
-        def main(value):
-            """:type value: typing.Optional[int]"""
-            return value
+                raise ValueError("Not nil")
+        def main(ints, strs):
+            """
+            :type ints: typing.List[typing.Optional[int]]
+            :type strs: typing.List[typing.Optional[str]]
+            """
+            return ints, strs
         self.assertEqual(
-            defopt.run(main, parsers={type(None): parser}, argv=[':none:']),
-            None)
+            defopt.run(main, parsers={type(None): parser},
+                       argv=['-i', 'nil', '0', '-s', 'nil', 's']),
+            ([None, 0], [None, 's']))
 
     def test_parse_bool(self):
         parser = defopt._get_parser(bool, {})
@@ -225,19 +360,19 @@ class TestParsers(unittest.TestCase):
         with self.assertRaisesRegex(Exception, 'no parser'):
             defopt._get_parser(object, parsers={type: type})
 
-    def test_list(self):
-        def main(foo):
-            """:type foo: list[float]"""
-            return foo
-        self.assertEqual(
-            defopt.run(main, argv=['--foo', '1.1', '2.2']), [1.1, 2.2])
+    def test_containers(self):
+        def main(foo, bar):
+            """
+            :type foo: tuple[float]
+            :type bar: list[float]
+            """
+            return foo, bar
+        self.assertEqual(defopt.run(main, argv=['1.1', '--bar', '2.2', '3.3']),
+                         ((1.1,), [2.2, 3.3]))
 
     def test_list_kwarg(self):
         def main(foo=None):
-            """Test function
-
-            :type foo: list[float]
-            """
+            """:type foo: list[float]"""
             return foo
         self.assertEqual(
             defopt.run(main, argv=['--foo', '1.1', '2.2']), [1.1, 2.2])
@@ -275,6 +410,82 @@ class TestParsers(unittest.TestCase):
         with self.assertRaises(SystemExit):
             defopt.run(main, argv=[])
 
+    def test_bool_optional(self):
+        def main(foo=None):
+            """:type foo: typing.Optional[bool]"""
+            return foo
+        def _parse_none(i):
+            if i.lower() == 'none':
+                return None
+            else:
+                raise ValueError('{} is not a valid None string'.format(i))
+        self.assertIs(defopt.run(main, argv=['1'],
+                      parsers={type(None): _parse_none}), True)
+        self.assertIs(defopt.run(main, argv=['0'],
+                      parsers={type(None): _parse_none}), False)
+        self.assertIs(defopt.run(main, argv=['None'],
+                      parsers={type(None): _parse_none}), None)
+        self.assertIs(defopt.run(main, argv=[],
+                      parsers={type(None): _parse_none}), None)
+
+    def test_bool_optional_keyword_none(self):
+        def main(*, foo=None):
+            """:type foo: typing.Optional[bool]"""
+            return foo
+        self.assertIs(defopt.run(main, argv=['--foo']), True)
+        self.assertIs(defopt.run(main, argv=['--no-foo']), False)
+        self.assertIs(defopt.run(main, argv=[]), None)
+
+    def test_bool_optional_keyword_true(self):
+        def main(*, foo=True):
+            """:type foo: typing.Optional[bool]"""
+            return foo
+        # cannot get a None value in this case from the CLI
+        self.assertIs(defopt.run(main, argv=['--foo']), True)
+        self.assertIs(defopt.run(main, argv=['--no-foo']), False)
+        self.assertIs(defopt.run(main, argv=[]), True)
+
+    def test_bool_optional_keyword_false(self):
+        def main(*, foo=False):
+            """:type foo: typing.Optional[bool]"""
+            return foo
+        self.assertIs(defopt.run(main, argv=['--foo']), True)
+        self.assertIs(defopt.run(main, argv=['--no-foo']), False)
+        self.assertIs(defopt.run(main, argv=[]), False)
+
+    def test_bool_optional_keyword_none_no_negated_flags(self):
+        def main(*, foo=None):
+            """:type foo: typing.Optional[bool]"""
+            return foo
+        self.assertIs(defopt.run(main, argv=['--foo'], no_negated_flags=True),
+                      True)
+        with self.assertRaises(SystemExit):
+            self.assertIs(defopt.run(main, argv=['--no-foo'],
+                          no_negated_flags=True), False)
+        self.assertIs(defopt.run(main, argv=[], no_negated_flags=True), None)
+
+    def test_bool_optional_keyword_true_no_negated_flags(self):
+        def main(*, foo=True):
+            """:type foo: typing.Optional[bool]"""
+            return foo
+        self.assertIs(defopt.run(main, argv=['--foo'], no_negated_flags=True),
+                      True)
+        # negated flag is still added, else foo could only be True
+        self.assertIs(defopt.run(main, argv=['--no-foo'],
+                      no_negated_flags=True), False)
+        self.assertIs(defopt.run(main, argv=[], no_negated_flags=True), True)
+
+    def test_bool_optional_keyword_false_no_negated_flags(self):
+        def main(*, foo=False):
+            """:type foo: typing.Optional[bool]"""
+            return foo
+        self.assertIs(defopt.run(main, argv=['--foo'], no_negated_flags=True),
+                      True)
+        with self.assertRaises(SystemExit):
+            self.assertIs(defopt.run(main, argv=['--no-foo'],
+                          no_negated_flags=True), False)
+        self.assertIs(defopt.run(main, argv=[], no_negated_flags=True), False)
+
     def test_bool_list(self):
         def main(foo):
             """:type foo: list[bool]"""
@@ -305,14 +516,26 @@ class TestParsers(unittest.TestCase):
         def main(foo='default'):
             """:type foo: bool"""
             return foo
-        self.assertIs(defopt.run(main, strict_kwonly=False,
+        self.assertIs(defopt.run(main, cli_options='has_default',
                                  argv=[]), 'default')
-        self.assertIs(defopt.run(main, strict_kwonly=False,
+        self.assertIs(defopt.run(main, cli_options='has_default',
                                  argv=['--foo']), True)
-        self.assertIs(defopt.run(main, strict_kwonly=False,
+        self.assertIs(defopt.run(main, cli_options='has_default',
                                  argv=['--no-foo']), False)
-        self.assertIs(defopt.run(main, strict_kwonly=False,
+        self.assertIs(defopt.run(main, cli_options='has_default',
                                  argv=['--foo', '--no-foo']), False)
+        with self.assertWarns(DeprecationWarning):
+            self.assertIs(defopt.run(main, strict_kwonly=False,
+                                     argv=[]), 'default')
+        with self.assertWarns(DeprecationWarning):
+            self.assertIs(defopt.run(main, strict_kwonly=False,
+                                     argv=['--foo']), True)
+        with self.assertWarns(DeprecationWarning):
+            self.assertIs(defopt.run(main, strict_kwonly=False,
+                                     argv=['--no-foo']), False)
+        with self.assertWarns(DeprecationWarning):
+            self.assertIs(defopt.run(main, strict_kwonly=False,
+                                     argv=['--foo', '--no-foo']), False)
 
     def test_bool_keyword_only(self):
         def main(*, foo):
@@ -323,6 +546,23 @@ class TestParsers(unittest.TestCase):
         with self.assertRaises(SystemExit):
             defopt.run(main, argv=[])
 
+    def test_cli_options(self):
+        def main(foo):
+            """:type foo: bool"""
+            return foo
+        self.assertIs(
+            defopt.run(main, cli_options='all', argv=['--foo']), True)
+        self.assertIs(
+            defopt.run(main, cli_options='all', argv=['--no-foo']), False)
+        with self.assertRaises(SystemExit):
+            defopt.run(main, cli_options='all', argv=['1'])
+        with self.assertRaises(SystemExit):
+            defopt.run(main, argv=['--foo'])
+        with self.assertRaises(SystemExit):
+            defopt.run(main, argv=['--no-foo'])
+        with self.assertRaises(SystemExit):
+            defopt.run(main, argv=[])
+
     def test_implicit_parser(self):
         def ok(foo):
             """:type foo: ConstructibleFromStr"""
@@ -344,7 +584,7 @@ class TestFlags(unittest.TestCase):
             """:type foo: int"""
             return foo
         self.assertEqual(
-            defopt.run(func, short={'foo': 'f'}, strict_kwonly=False,
+            defopt.run(func, short={'foo': 'f'}, cli_options='has_default',
                        argv=['-f', '2']),
             2)
 
@@ -375,6 +615,12 @@ class TestFlags(unittest.TestCase):
         def func(*, hello="world"):
             """:type hello: str"""
         defopt.run(func, argv=[])
+        with self.assertRaises(SystemExit):
+            defopt.run(func, argv=["-h", ""])
+        self.assertEqual(
+            defopt.run(
+                func, argparse_kwargs={"add_help": False}, argv=["-h", ""]),
+            None)
 
 
 class TestEnums(unittest.TestCase):
@@ -414,9 +660,14 @@ class TestTuple(unittest.TestCase):
             return foo
         self.assertEqual(defopt.run(main, argv=['1', '2']), (1, '2'))
 
-    def test_tupleenum(self):
+    def test_tuple_variable_length(self):
         def main(foo):
-            """:param typing.Tuple[Choice] foo: foo"""
+            """:param typing.Tuple[int,...] foo: foo"""
+            return foo
+        self.assertEqual(defopt.run(main, argv=['1', '2', '3']), (1, 2, 3))
+
+    def test_tupleenum(self):
+        def main(foo: typing.Tuple[Choice] = None):
             return foo
         self.assertEqual(defopt.run(main, argv=['one']), (Choice.one,))
 
@@ -434,6 +685,12 @@ class TestTuple(unittest.TestCase):
         self.assertEqual(str(defopt.run(main, argv=['1', '2', '3'])),
                          str(Pair(1, '2')))
 
+    def test_tuple_fails_early(self):
+        def main(foo):
+            """:param typing.Tuple[int,EmptyType] foo: foo"""
+        with self.assertRaisesRegex(Exception, 'no parser'):
+            defopt.run(main, argv=['-h'])
+
 
 class TestUnion(unittest.TestCase):
     def test_union(self):
@@ -448,6 +705,15 @@ class TestUnion(unittest.TestCase):
         self.assertEqual(defopt.run(main, argv=['1', 'b']), (int, str))
         self.assertEqual(defopt.run(main, argv=['a', '2']), (str, float))
 
+    def test_or_union(self):
+        if not hasattr(types, "UnionType"):
+            raise unittest.SkipTest("A|B-style unions not supported")
+        def main(foo):
+            """:param int|str foo: foo"""
+            return type(foo)
+        self.assertEqual(defopt.run(main, argv=['1']), int)
+        self.assertEqual(defopt.run(main, argv=['x']), str)
+
     def test_bad_parse(self):
         def main(foo):
             """:param typing.Union[int,float] foo: foo"""
@@ -512,7 +778,7 @@ class TestDoc(unittest.TestCase):
         :junk one two: nothing
         """
         doc = defopt._parse_docstring(inspect.cleandoc(doc))
-        self.assertEqual(doc.text, 'Test function')
+        self.assertEqual(doc.text, 'Test function\n\n')
         one = doc.params['one']
         self.assertEqual(one.text, 'first param')
         self.assertEqual(one.type, 'int')
@@ -571,11 +837,11 @@ class TestDoc(unittest.TestCase):
 
     def test_implicit_role(self):
         doc = defopt._parse_docstring("""start `int` end""")
-        self.assertEqual(doc.text, 'start \033[4mint\033[0m end')
+        self.assertEqual(doc.text, 'start \033[4mint\033[0m end\n\n')
 
     def test_explicit_role(self):
         doc = defopt._parse_docstring("""start :py:class:`int` end""")
-        self.assertEqual(doc.text, 'start int end')
+        self.assertEqual(doc.text, 'start int end\n\n')
 
     def test_sphinx(self):
         doc = """
@@ -651,10 +917,10 @@ class TestDoc(unittest.TestCase):
         self._check_doc(doc)
 
     def _check_doc(self, doc):
-        self.assertRegex(
+        self.assertEqual(
             doc.text,
-            r'\AOne line summary\. \n \nExtended description\.( \n)+'
-            r'examples: \n>>> print\("hello, world"\)\Z')
+            'One line summary.\n\nExtended description.\n\n'
+            'examples:\n\n>>> print("hello, world")\n\n')
         self.assertEqual(len(doc.params), 3)
         self.assertEqual(doc.params['arg1'].text, 'Description of arg1')
         self.assertEqual(doc.params['arg1'].type, 'int')
@@ -669,25 +935,35 @@ class TestDoc(unittest.TestCase):
             defopt._get_type_from_doc('Sequence[int]', globalns),
             typing.List[int])
 
+    def test_collection(self):
+        # This test effectively does nothing on python versions < 3.6
+        globalns = {'Collection': getattr(typing, 'Collection', typing.List)}
+        self.assertEqual(
+            defopt._get_type_from_doc('Collection[int]', globalns),
+            typing.List[int])
+
     def test_iterable(self):
         globalns = {'typing': typing}
         self.assertEqual(
             defopt._get_type_from_doc('typing.Iterable[int]', globalns),
             typing.List[int])
 
-    def test_other(self):
-        with self.assertRaisesRegex(ValueError, 'unsupported.*tuple'):
-            defopt._get_type_from_doc('tuple[int]', {})
-
     def test_literal_block(self):
         doc = """
         ::
 
             Literal block
                 Multiple lines
+
+        .. code:: python
+
+            def foo(): pass
         """
         doc = defopt._parse_docstring(inspect.cleandoc(doc))
-        self.assertEqual(doc.text, '    Literal block\n        Multiple lines')
+        self.assertEqual(
+            doc.text,
+            '    Literal block\n        Multiple lines\n\n'
+            '    def foo(): pass\n\n')
 
     def test_newlines(self):
         doc = """
@@ -711,26 +987,25 @@ class TestDoc(unittest.TestCase):
         #.  baz
         """
         doc = defopt._parse_docstring(inspect.cleandoc(doc))
-        # Use inspect.cleandoc and not textwrap.dedent, as we want to keep
-        # whitespace in lines that contain more than the common leading
-        # whitespace.
         self.assertEqual(doc.text, inspect.cleandoc("""\
             Bar
-            Baz 
-             
-            - bar 
-            - baz 
-             
-            quux: 
-             
-                hello 
-             
-             
-            1. bar 
-            2. baz 
-             
-            ii.  bar 
-            iii. baz"""))
+            Baz
+
+            - bar
+
+            - baz
+
+            quux:
+
+                hello
+
+            1. bar
+
+            2. baz
+
+            ii.  bar
+
+            iii. baz""") + "\n\n")
 
 
 class TestAnnotations(unittest.TestCase):
@@ -912,7 +1187,7 @@ class TestHelp(unittest.TestCase):
         self.assertLessEqual({*flags}, {'d', 't', 'n'})
         parser = defopt._create_parser(
             funcs, show_defaults='d' in flags, show_types='t' in flags,
-            no_negated_flags='n' in flags, strict_kwonly=False)
+            no_negated_flags='n' in flags, cli_options='has_default')
         return parser.format_help()
 
 
@@ -941,6 +1216,12 @@ class TestVersion(unittest.TestCase):
              self._assert_streams(stdout=r'\Afoo 42\n\Z'):
             defopt.run([], version='foo 42', argv=['--version'])
 
+    def test_moduleless(self):
+        def moduleless(): pass
+        moduleless.__module__ = None
+        with self.assertRaises(ValueError):
+            defopt.run([moduleless], version=True)
+
     @contextlib.contextmanager
     def _assert_streams(self, *, stdout=None, stderr=None):
         with ExitStack() as stack:
@@ -1136,3 +1417,12 @@ class TestDefaultsPreserved(unittest.TestCase):
         params = inspect.signature(defopt.run).parameters
         self.assertEqual(params['parsers'].default, {})
         self.assertEqual(params['argparse_kwargs'].default, {})
+
+
+class TestStyle(unittest.TestCase):
+    def test_line_length(self):
+        for path in [defopt.__file__, __file__]:
+            with tokenize.open(path) as src:
+                for i, line in enumerate(src, 1):
+                    if len(line) > 80:  # 79 chars + newline.
+                        self.fail('{}:{} is too long'.format(path, i))
